/* 
 * Zeus URL rewriting ISAPI filter
 * 
 * This short ISAPI filter rewrites incoming URLs
 * into something different by calling an external
 * process. This process is long-lived and sits at
 * the end of a socketpair(). The zeus.web webserver
 * process will then feed URLs as HTTP requests 
 * arrive from users' browsers through the socket
 * and redirect the browsers whereever the external
 * process directs them.
 *
 * If the URL is the same, no redirection is 
 * performed. Care should be taken that the external
 * process does not block, else the entire webserver
 * will stop serving requests, due to the single-threaded
 * nature of the Zeus Web Server.
 *
 * (c) Jon Thackray, 4th February 2001.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>

#include "wintypes.h"
#include "httpfilt.h"

/* Global (necessary) variables */
static char* rewriter_prog = "/usr/local/zeus/rewriter";
static int initdone = 0;
static pid_t child = 0;
static int fd[2];


DWORD HttpFilterProc( HTTP_FILTER_CONTEXT *pfc,
                      DWORD NotificationType,
                      VOID* notificationInfo )
{
   char buff[256];
   char buff2[256];
   char url[256];
   int  urllen = 256;
   int s;

   PHTTP_FILTER_PREPROC_HEADERS headers = notificationInfo;

   url[0] = '\0';

   /* get url from headers */
   headers->GetHeader( pfc, "url", url, &urllen );

   fprintf(stderr, "jrmt: old url = %s\n", url);
   write( fd[0], url, urllen );
   s = read( fd[0], buff, sizeof(buff) );
   buff[s] = '\0';
   fprintf(stderr, "jrmt: new url = %s\n", buff);
   sprintf( buff2, "Location: %s\n", buff );

   /* headers->SetHeader( pfc, "url", buff ); */

   pfc->ServerSupportFunction( pfc,
                               SF_REQ_SEND_RESPONSE_HEADER,
                               (PVOID) "302 Over There(tm)",
                               (PVOID) buff2,
                               0);

   /* do the job */
   return SF_STATUS_REQ_NEXT_NOTIFICATION;
}


/* Termination, doesn't get called if filter is unloaded */
extern BOOL TerminateFilter( DWORD dwFlags )
{
   /* do the job */
   fprintf( stderr, "stopping pid %d\n", child );
   kill( child, 9 );
   return TRUE;
}


/* Initialisation */
BOOL GetFilterVersion( HTTP_FILTER_VERSION *pVer )
{
   struct stat sbuf;

   pVer->dwFilterVersion = HTTP_FILTER_REVISION;
   strncpy( pVer->lpszFilterDesc, "External process URL rewriter", SF_MAX_FILTER_DESC_LEN );

   /* Ask to be notified, when the headers have been processed */
   pVer->dwFlags = SF_NOTIFY_PREPROC_HEADERS;

   /* Firstly, try to startup an external rewriter process
    * by fork()ing and keeping one end of a pipe to the stdin
    * of the rewriter process, and stdout will contain the
    * newly mapped URL... 
    */

   if ( stat( rewriter_prog, &sbuf ) == -1 ) {
      fprintf( stderr, "%s cannot be stat()ed, exiting\n", rewriter_prog );
      return FALSE;
   }

   /* More paranoia, don't run if writable by others */
   if ( ! S_ISREG(sbuf.st_mode)
        || (sbuf.st_mode & S_IXUSR) != S_IXUSR
        || (sbuf.st_mode & S_IWGRP) == S_IWGRP
        || (sbuf.st_mode & S_IWOTH) == S_IWOTH ) {
      fprintf( stderr, "%s not executable or non-owner writable, exiting\n", rewriter_prog );
      return FALSE;
   }

   if ( socketpair( AF_UNIX, SOCK_STREAM, 0, fd ) == -1 ) {
      fprintf( stderr, "cannot create socketpair()\n" );
      return FALSE;
   }

   /* Once fork()ed, parent reads/writes to fd0
    * and child reads/writes to fd1. Make child
    * stdin and stdout our end of the socketpair
    */
   child = fork();
   if ( child == 0 ) {
      char* env[1];
      close( fd[0] );
      env[0] = NULL;
      dup2( fd[1], 0 );
      dup2( fd[1], 1 );

      if ( execle( rewriter_prog, rewriter_prog, NULL, env ) == -1 ) {
         fprintf( stderr, "cannot exec() %s because %s\n", rewriter_prog, strerror( errno ) );
         initdone = 0;
         exit( 0 );
      }
      /* Should never happen, but be paranoid :) */
      exit( 0 );
   } else {
      close( fd[1] );
      initdone = 1;
   }
 
   fprintf( stderr, "successfully fork()ed %s as pid %d\n", rewriter_prog, child );
   return TRUE;
}

